/* 
   DDS, a bridge double dummy solver.

   Copyright (C) 2006-2014 by Bo Haglund / 
   2014-2018 by Bo Haglund & Soren Hein.

   See LICENSE and README.
*/


#ifndef DTEST_ARGS_H
#define DTEST_ARGS_H

void Usage(
 const char base[]);

void PrintOptions();

void ReadArgs(
  int argc,
  char * argv[]);

#endif

